@@ -1,6 +1,6 @@
-// 2004-01-11  Petur Runolfsson  <peturr02@ru.is>
+// 2004-04-16  Petur Runolfsson  <peturr02@ru.is>
 
-// Copyright (C) 2004 Free Software Foundation, Inc.
+// Copyright (C) 2004, 2005 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
@@ -18,51 +18,64 @@
 // Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 // USA.
 
-// 27.8.1.4 Overridden virtual functions
-
 #include <fstream>
 #include <locale>
+#include <unistd.h>
+#include <signal.h>
+#include <fcntl.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <testsuite_hooks.h>
 
-// libstdc++/13582
+// libstdc++/14975
 void test01()
 {
-  bool test __attribute__((unused)) = true;
   using namespace std;
+  using namespace __gnu_test;
+  bool test __attribute__((unused)) = true;
 
-  locale loc_en(__gnu_test::try_named_locale("en_US"));
-  locale loc_fr(__gnu_test::try_named_locale("fr_FR"));
+  locale loc_us = try_named_locale("en_US");
 
-  const char* name = "tmp_13582-3.tst";
+  const char* name = "tmp_14975-2";
 
-  {
-    filebuf fbout;
-    fbout.open(name, ios_base::out);
-    fbout.sputn("AbCdE", 5);
-    fbout.close();
-  }
+  signal(SIGPIPE, SIG_IGN);
 
-  {
-    wfilebuf fbin;
-    fbin.open(name, ios_base::in);
-    wfilebuf::int_type n = fbin.sbumpc();
-    VERIFY( n == L'A' );
-    fbin.pubimbue(loc_en);
-    fbin.pubseekoff(0, ios_base::cur);
-    n = fbin.sbumpc();
-    VERIFY( n == L'b' );
-    fbin.pubimbue(loc_fr);
-    n = fbin.sbumpc();
-    VERIFY( n == L'C' );
-    n = fbin.sbumpc();
-    VERIFY( n == L'd' );
-    fbin.pubseekoff(0, ios_base::cur);
-    n = fbin.sbumpc();
-    VERIFY( n == L'E' );
-    n = fbin.sbumpc();
-    VERIFY( n == wfilebuf::traits_type::eof() );
-    fbin.close();
-  }
+  unlink(name);  
+  try_mkfifo(name, S_IRWXU);
+  semaphore s1;
+
+  int child = fork();
+  VERIFY( child != -1 );
+
+  if (child == 0)
+    {
+      {
+	filebuf fbin;
+	fbin.open(name, ios_base::in);
+      }
+      s1.signal ();
+      exit(0);
+    }
+  
+  wfilebuf fb;
+  fb.pubimbue(loc_us);
+  wfilebuf* ret = fb.open(name, ios_base::out);
+  VERIFY( ret != NULL );
+  VERIFY( fb.is_open() );
+
+  s1.wait ();
+
+  try
+    {
+      fb.sputc(L'a');
+      fb.sputc(L'b');
+      fb.pubimbue(locale::classic());
+      fb.sputc(L'c');
+      fb.close();
+    }
+  catch (std::exception&)
+    {
+    }
 }
 
 int main()
