@@ -26,10 +26,8 @@
 int main()
 {
   using namespace __gnu_test;
-
- // cplus-dem FAIL
-  verify_demangle("_Z3fooA30_A_i", "foo(int [30][])");
-
+  // cplus-dem FAIL
+  verify_demangle("_Z1fA37_iPS_", "f(int [37], int (*) [37])");
 
   return 0;
 }
